/************************************************************************/
/*                                                                      */
/*             ,     ..        */
/*                   ,   ,   .               */
/*                                                                      */
/************************************************************************/

/************************************************************************/
/* IR receiver support implementation                                   */
/************************************************************************/

#include <htc.h>
#include "Include\PIC_Common.h"
#include "ir.h"

// Oscillator frequency
#define F_OSC					20000000


// Function prototypes
static void IR_RcvInit(void);
static void IR_RcvReset(void);
static void IR_RcvResetTimer(char scale);
static void IR_RcvOnInt(void);
static char IR_TimeInBounds(TimeBounds bounds);
static char IR_RcvPutBit(char cbit);
static void IR_RcvPutCodeToBuffer(void);

// Global IR variables
IR_RcvConfig g_rcvcfg;
unsigned long g_rcvcode = 0;
char g_iscode = 0;

// Internal IR variables
static IR_ReceiverState g_rcvst;
static char g_intedg = 0;
static unsigned long g_ircode;
static char g_bitcount;


static long IR_TimeToCode(long time, char scale)
{
	return time * (F_OSC / 1000l) / ((1l << (scale + 1)) * 4000l);
}

void IR_SetDefConfig(void)
{
	g_rcvcfg.data_bits			= RCV_DATA_BITS;
	g_rcvcfg.hd_tmr_scale		= 7;
	g_rcvcfg.hd_space_time.tmin	= IR_TimeToCode(RCV_DEF_HD_SPACE_TIME_MIN, g_rcvcfg.hd_tmr_scale);
	g_rcvcfg.hd_space_time.tmax	= IR_TimeToCode(RCV_DEF_HD_SPACE_TIME_MAX, g_rcvcfg.hd_tmr_scale);
	g_rcvcfg.hd_mark_time.tmin	= IR_TimeToCode(RCV_DEF_HD_MARK_TIME_MIN, g_rcvcfg.hd_tmr_scale);
	g_rcvcfg.hd_mark_time.tmax	= IR_TimeToCode(RCV_DEF_HD_MARK_TIME_MAX, g_rcvcfg.hd_tmr_scale);
	g_rcvcfg.dt_tmr_scale		= 5;
	g_rcvcfg.dt_pause_time.tmin	= IR_TimeToCode(RCV_DEF_DT_PAUSE_TIME_MIN, g_rcvcfg.dt_tmr_scale);
	g_rcvcfg.dt_pause_time.tmax	= IR_TimeToCode(RCV_DEF_DT_PAUSE_TIME_MAX, g_rcvcfg.dt_tmr_scale);
	g_rcvcfg.dt_0_time.tmin		= IR_TimeToCode(RCV_DEF_DT_0_TIME_MIN, g_rcvcfg.dt_tmr_scale);
	g_rcvcfg.dt_0_time.tmax		= IR_TimeToCode(RCV_DEF_DT_0_TIME_MAX, g_rcvcfg.dt_tmr_scale);
	g_rcvcfg.dt_1_time.tmin		= IR_TimeToCode(RCV_DEF_DT_1_TIME_MIN, g_rcvcfg.dt_tmr_scale);
	g_rcvcfg.dt_1_time.tmax		= IR_TimeToCode(RCV_DEF_DT_1_TIME_MAX, g_rcvcfg.dt_tmr_scale);
}

// Initialization
void IR_Init(void)
{
	IR_RcvInit();

	// Initialize Timer1
	T1CON = 0;			// set timer mode and prescaler (1:1)
	TMR1IE = 0;			// disable timer interrupt
	TMR1IF = 0;
}

// Receiver routines

static void IR_RcvInit(void)
{
	T0IE = 0;
	INTE = 0;
	T0CS = 0;
	PSA = 0;
	INTEDG = g_intedg;
	g_rcvst = rcvstIdle;
	INTE = 1;
}

static void IR_RcvReset(void)
{
	// Set IR receiver idle state
	T0IE = 0;
	T0IF = 0;
	INTF = 0;
	g_ircode = 0;
	g_bitcount = 0;
	g_rcvst = rcvstIdle;
}

static void IR_RcvResetTimer(char scale)
{
	T0IE = 0;
	PS0 = scale & 1;
	PS1 = (scale >> 1) & 1;
	PS2 = (scale >> 2) & 1;
	TMR0 = 0;
	T0IF = 0;
	T0IE = 1;
}

static void IR_RcvOnInt(void)
{
	char cbit;
	
	switch (g_rcvst)
	{
		case rcvstIdle:
			IR_RcvResetTimer(g_rcvcfg.hd_tmr_scale);
			g_rcvst = rcvstHdSpace;
			break;
			
		case rcvstHdSpace:
			if (!IR_TimeInBounds(g_rcvcfg.hd_space_time))
			{
				IR_RcvReset();
			}
			else
			{
				IR_RcvResetTimer(g_rcvcfg.hd_tmr_scale);
				g_rcvst = rcvstHdMark;
			}
			break;
			
		case rcvstHdMark:
			if (!IR_TimeInBounds(g_rcvcfg.hd_mark_time))
			{
				IR_RcvReset();
			}
			else
			{
				IR_RcvResetTimer(g_rcvcfg.dt_tmr_scale);
				g_rcvst = rcvstDtSpace;
			}
			break;
			
		case rcvstDtSpace:
			if (!IR_TimeInBounds(g_rcvcfg.dt_pause_time))
			{
				IR_RcvReset();
			}
			else
			{
				IR_RcvResetTimer(g_rcvcfg.dt_tmr_scale);
				g_rcvst = rcvstDtMark;
			}
			break;

		case rcvstDtMark:
			if (IR_TimeInBounds(g_rcvcfg.dt_1_time))
			{
				cbit = 1;
			}
			else if (IR_TimeInBounds(g_rcvcfg.dt_0_time))
			{
				cbit = 0;
			}
			else
			{
				cbit = 2;
				IR_RcvReset();
			}
			if (cbit != 2)
			{
				IR_RcvResetTimer(g_rcvcfg.dt_tmr_scale);
				if (IR_RcvPutBit(cbit))
				{
					g_rcvst = rcvstTerm;
				}
				else
				{
					g_rcvst = rcvstDtSpace;
				}
			}
			break;
			
		case rcvstTerm:
			if (IR_TimeInBounds(g_rcvcfg.dt_pause_time))
			{
				IR_RcvPutCodeToBuffer();
			}
			IR_RcvReset();
			break;
	}
}

static char IR_TimeInBounds(TimeBounds bounds)
{
	char tmr0 = TMR0;
	return (tmr0 < bounds.tmin || tmr0 > bounds.tmax) ? 0 : 1;
}

static char IR_RcvPutBit(char cbit)
{
	g_ircode = (g_ircode << 1) | cbit;
	return (++g_bitcount >= RCV_DATA_BITS) ? 1 : 0;
}

static void IR_RcvPutCodeToBuffer(void)
{
	g_rcvcode = g_ircode;
	g_iscode = 1;
}

void IR_HandleInterrupt()
{
	// Timer0 interrupt
	if (g_rcvst != rcvstIdle && T0IF)
	{
		// Timeout - set IR receiver idle state
		IR_RcvReset();
	}
	// TSOP interrupt
	else if (INTF)
	{
		// Switch interrupt polarity
		g_intedg ^= 1;
		INTEDG = g_intedg;			
		INTF = 0;
		// Handle state change
		IR_RcvOnInt();
	}
}
